////////////////////////////////////////////////////////////////
// Hartmut Ring
// Capella 2008, Version 6.0 - Dokumentation des Dateiformats
//
// Stand vom 18.07.2007
//
// +---------------------------------------------------------+
// |  Wo sich die Formate von capella 2000 (Version 3),      |
// |  capella 2002 (Version 4), capella 2004 (Version 5)     |
// |  und capella 2008 (Version 6)  unterscheiden,           |
// |  ist dies kenntlich gemacht mit:                        |
// |  #ifdef _VER_3_ ... bzw.                                |
// |  #ifdef _VER_4_ ... bzw.                                |
// |  #ifdef _VER_5_ ... bzw.                                |
// |  #ifdef _VER6_EXT_ ...                                  |
// ----------------------------------------------------------+
//
// Um Fehler und Missverstndnisse auszuschlieen, die sich bei
// einer verbalen Beschreibung des Dateiformats nicht vermeiden
// lassen, hat diese Dokumentation die Form eines einfachen
// kommentierten C++-Programms (32-bit Windows-EXE), das eine
// capella-Datei einliest, ohne die Daten auszuwerten.
// Um die Lesbarkeit zu erhhen, ist das Rahmenprogramm mit allen
// Dateizugriffen in die Datei "LowLevel.cpp" ausgelagert.
//
// Jede Definition beginnt mit einer Zeile der Form
// DEFINITION (Typ)
// Dahinter verbirgt sich die Implementierung des Konstruktors,
// der ein Objekt dieses Typs aus einer Datei einliest.
//
// ----------------------------------------------------
// zum Studium des Dateiformats beginnen Sie bitte bei
// DEFINITION (Score)
// Alle anderen Definitionen sind alphabetisch geordnet
// ----------------------------------------------------
//
// Die Original-Implementierung von capella 2002 verwendet polymorphe
// Klassenhierarchien (z. B. fr Notenobjekte und Grafikobjekte).
// Um dieses "Dokumentationsprogramm" mglichst kompakt zu halten,
// wurde stattdessen eine vereinfachte Darstellung gewhlt.
//

#define TRACE
#ifdef TRACE
 #include <stdio.h>
#endif

//#define _VER_3_
#define _VER_5_
// sonst: Version 4
#define _VER6_EXT_

////////////////////////////////////////////////////////////////
#include "CapRead.h"

////////////////////////////////////////////////////////////////

/* allgemeines Layout (zum Ausdruck: lngste Zeile 90 Zeichen!)

  +-------------+-------------------+---------------------------------
  Typ           erklrender Name    Kommentar
  +-------------+-------------------+---------------------------------
  |             |                   teilweise in Form von Anweisungen,
  |             |                   die den Wertebereich einschrnken (assert)
  |             |                   oder Teildaten (Bit-Bereiche) auswerten
  |             |                   |
  |             |                   |
  :             :                   :
*/


#ifdef _VER_3_
#else
// die folgenden Variablen mussten leider global deklariert werden,
// weil in DEFINITION(Voice) fr jedes Notenobjekt Extra-Informationen
// ermittelt werden, die erst im konkreten Objekt ausgewertet werden.
static int  glb_nExtraBytes;
static BYTE glb_extraData[200];
#endif


DEFINITION(Score) {
  CharArray     header(8);          // Dateikennung (8 Bytes)
                                    assert(memcmp (header, "cap3-v:", 7) == 0);
                                    assert(header[7] >= 'a' && header[7] <= 'z');
                                    // derzeitige Version: header[7] == 'a'
                                    // Es sollen aber auch zuknftige Versionen akzeptiert
                                    // werden, wobei die Zusatzinformationen zwar gelesen
                                    // aber nicht interpretiert werden.
  String        author;
  String        keywords;
  String        comment;

  Unsigned      nRel;               // Darstellungsparameter
  Unsigned      nAbs;               // dto.
  Byte          b;                  bool bUseRealSize      = (b & 1) != 0;
                                    bool bAllowCompression = (b & 2) != 0;
                                    bool bPrintLandscape   = (b & 16) != 0; // Querformat
                                    assert((b & 0xEC) == 0);  // bits 2,3,5,6,7 reserviert

  ScoreLayout   layout;             // Mustersystem
  Byte          beamRelMin0;        // Grundeinstellungen fr Balkensteigung
  Byte          beamRelMin1;
  Byte          beamRelMax0;
  Byte          beamRelMax1;
#ifdef _VER_3_
  ExtraInfo     extra;
#else
  // neu in Version 4.0:
  // wahlweise explizite MIDI-Kanalzuordnung
  Unsigned      nExtraBytes;
  if (nExtraBytes > 0) {            int nRead = 0;
    while (nRead < nExtraBytes) {
      Byte      type;               nRead++;
      if (type == 1) {              // einziges in Version 4 bekanntes Format
        Byte    n;                  nRead++; // Anzahl der Zeilen im Mustersystem
        FOR (n)
          Byte  preferredChannel;   nRead += n;
      }
 #ifdef _VER_5_
      else if (type == 2) {         // ab Version 5: 
        Byte    n;                  nRead++;
                                    assert(n == 2);
        Word pageObjLineDist;       // Notenlinienabstand fr die virtuelle Notenzeile,
                                    // an der die Seitenobjekte verankert sind.
                                    nRead += 2;
      }
 #endif
#ifdef _VER6_EXT_
      else if (type == 3) {         // ab Version 5: 
        // Taktnummerierung: Zusatz-Info
        Byte    n;                  nRead++;
                                    assert(n == 13);
        Byte showBarCountInterval;
        Byte barCountInterval;
        Byte barCountFill;
        Byte barCountLineWidth;
        Byte(barCountFillOpacity);

        union {
          COLORREF color[2];        // Rand-/Fllfarbe
          BYTE b[8];
        };
        for (int i=0; i<8; i++) {
          Byte d;
          b[i] = d;
        }
                                    COLORREF barNumPenColor = color[0];
                                    COLORREF barNumBrushColor = color[1];
                                    nRead += 13;
      }
#endif
      else {                        // in Version 4 unbekanntes Format
        Byte    n;                  nRead++;
        FOR (n)
          Byte  unknown(n);        nRead += n;
      }
      assert(nRead <= nExtraBytes);
    }
    if (nRead < nExtraBytes) {
      int d = nExtraBytes-nRead;
      FOR (d)
        Byte    unknown;
    }
  }
#endif

  DrawObjArray  objects;            // Galerie (gesammelte Grafikobjekte)
  StringArray   objectNames;        // Namen der Galerie-Objekte
                                    assert(objects.Size() == objectNames.Size());

  ChordObj      backgroundChord;    // enthlt die Grafikobjekte
                                    // auf dem Seitenhintergrund

  Bool          bShowBarCount;      // Taktnumerierung zeigen
  Byte          barNumberFrame;     // 0=kein, 1=Rechteck, 2=Ellipse
#ifdef _VER6_EXT_  
  Byte          nBarDistX;          // 0..32 --> 0..32, 33..64 --> -1..-32
#else
  Byte          nBarDistX;          // (positiv)
#endif
  Byte          nBarDistY;          // dto.
  LogFont       barNumFont;

  Unsigned      nFirstPage;         // Versatz fr Seitenzhlung

  Unsigned      leftPageMargins;    // Seitenrnder
  Unsigned      topPageMargins;
  Unsigned      rightPageMargins;
  Unsigned      btmPageMargins;

  Unsigned      nSystems;
  FOR (nSystems)
    System      system;             // Systeme

  CharArray     reserved(4);        // Am Dateiende 4 Bytes reserviert
                                    assert(memcmp(reserved, "\0\0\0\0", 4) == 0);
}

////////////////////////////////////////////////////////////////
// Die folgenden Basistypen sind in LowLevel.cpp definiert:

// - Byte      (1 Byte)
// - Char      (1 Byte)
// - Bool      (1 Byte)

// - Word      (2 Bytes)
// - DWord     (4 Bytes)
// - Integer   (flexibel: 1-5 Bytes)
// - Unsigned  (flexibel: 1-5 Bytes)
// - Color     (flexibel: 1-4 Bytes)

// - CharArray
// - String

////////////////////////////////////////////////////////////////
// Die folgenden Typen sind alphabetisch angeordnet:

DEFINITION(BasicDrawObj) {
  Byte          modeX;              // anchorMode
  Byte          modeY;
  Byte          distY;
  Byte          flags;
  Integer       nRefNote;
  Word          range;              // bit 0..11  Notenbereich
                                    int nNotes = range & 0x0FFF;
                                    // bit 12     Hinter Noten
                                    bool background = (range & 0x1000) != 0;
                                    // bit 13..15 Seitenbereich
                                    int pageRange = (range >> 13) & 7;
}

DEFINITION(BasicDurationalObj) {
  Byte          b;                  int  nDots =  b & 0x03;
                                    bool noDuration = (b & 0x04) != 0;
                                    bool postGrace  = (b & 0x08) != 0; // Nachschlag
                                    bool bSmall     = (b & 0x10) != 0;
                                    bool invisible  = (b & 0x20) != 0;
                                    bool notBlack   = (b & 0x40) != 0;
                                    assert((b & 0x80) == 0); // bit 7 reserviert
  if (b & 0x40)
    Color       color;              // sonst schwarz

  Byte          c;                  // Notenwert
                                    enum TIMESTEP {D1, D2, D4, D8, D16, D32,
                                                   D64, D128, D256, D_BREVE};
                                    TIMESTEP t = (TIMESTEP) (c & 0x0F); // duration step
  if (c & 0x10)
    Integer     horizontalShift;    // Horizontalverschiebung

  if (c & 0x20) {
    Byte        tuplet;             // irregulre Teilung
                                    int  count        =  tuplet & 0x0F;
                                    bool tripartite   = (tuplet & 0x10) != 0;
                                    bool isProlonging = (tuplet & 0x20) != 0;
                                    assert((tuplet & 0xC0) == 0); // bits 6,7 reserviert
  }

  if (c & 0x40) {
    DrawObjArray objects;           // Grafikobjekte
                                    assert(objects.Size() > 0);
  }
                                    assert((c & 0x80) == 0); // bit 7 reserviert
}

DEFINITION(BasicRectObj) {
  BasicDrawObj  base;
  Point         relPos;
  Integer       width;
  Integer       yxRatio;            int height = MulDiv(width, yxRatio, 0x10000);
}

DEFINITION(BracketObj) {
  LineObj       line;
  Byte          orientation;
  Byte          number;
}

DEFINITION(ChordObj) {
  BasicDurationalObj base;

  Byte          flags;
  if (flags & 0x01)                 // sonst AUTO_BEAM
    Byte        beamMode;               // 1=FORCE_BEAM, 2=SPLIT_BEAM

  if (flags & 0x02) {
    Char        notationStave;      assert(notationStave >= -1
                                        && notationStave <= 1);
  }

  if (flags & 0x04) {
    Char        stemDir;            assert(stemDir == -1   // ab
                                        || stemDir == 0    // auto
                                        || stemDir == 1    // auf
                                        || stemDir == 3);  // kein
    Char        dStemLength;
  }                                 // sonst beides Null

  if (flags & 0x08) {
    Byte        nTremoloBars;       assert(nTremoloBars <= 5);
    Byte        articulation;       assert(articulation >= 0
                                        && articulation <= 8);
  }

  if (flags & 0x10) {
    Byte        b;                  // Haltebogen
                                    bool leftTie  = (b & 1) != 0;
                                    bool rightTie = (b & 2) != 0;
  }

  if (flags & 0x20) {
    Char        beamShift;
    Char        beamSlope;
  }

  if (flags & 0x40) {               // Liedtext
    Unsigned    nVerses;
    FOR (nVerses) {
      Bool      bVerse;
      if (bVerse) {                 // Liedtextsilbe der i-ten Strophe
        Byte    b;                  bool leftAlign = (b & 1) != 0;
                                    bool extender  = (b & 2) != 0;
                                    bool hyphen    = (b & 4) != 0;
        if (b & 8)
          String verseNumber;
        if (b & 16)
          String text;
      }
    }
  }

  Unsigned      nNotes;             assert(nNotes > 0);
  FOR (nNotes) {                    // note head data
    // Die Tonhhe war in frheren Entwicklungsversionen in einem signed char gespeichert.
    // Der Wert lag im Bereich -64 .. 63.
    // Also waren folgende Bytes mglich:
    // 11xxxxxx -1 .. -64
    // 00xxxxxx  0 ..  63
    // Um mit den alten Dateien kompatibel zu bleiben, wird 
    // in der neuen Version ein erzwungenes Vorzeichen durch ein
    // vom zweithchsten Bit abweichendes hchstes Bit markiert:
    Char        c;                  bool bit7 = (c & 0x80) != 0;
                                    bool bit6 = (c & 0x40) != 0;
                                    int explAlteration = 0;
                                    char pitch = c;
                                    // Diatonischer Notenwert relativ zu aktueller Tonart
                                    // und Schlssel.
                                    // z. B. 0 = Ausgangston, 7 = eine Oktave hher
                                    if (bit7 != bit6) {
                                      explAlteration = 2; // Vorzeichen unterdrcken
                                      pitch ^= 0x80; // Bit 7 an Bit 6 angleichen
                                    }
    Byte        b;                  int headType = b & 7;  // 0=normaler Kopf
                                    assert(headType >= 0 && headType <= 6);

                                    int alteration = ((b >> 3) & 7) - 2;
                                    assert(alteration >= -2 && alteration <= 2);

                                    if ((b & 0x40) != 0) {
                                      // nicht erzwingen und unterdrcken gleichzeitig:
                                      assert(explAlteration == 0);
                                      explAlteration = 1; // Vorzeichen erzwingen
                                    }
                                    bool silent = (b & 0x80) != 0;
  }
#ifdef _VER_3_
#else
  // Neue Attribute in capella 4 aus den in DEFINITION(Voice) global abgelegten
  // Zusatzinformationen (glb_nExtraBytes, glb_extraData) holen:
  if (glb_nExtraBytes > 0) {
    assert(glb_nExtraBytes > 1);
    assert(glb_extraData[0] == 1); // Kennung fr capella-4-Format
    int i = 1;
    while (i < glb_nExtraBytes-1) {
      BYTE format = glb_extraData[i++];
      BYTE nBytes = glb_extraData[i++];

      switch (format) {
      // Notenkopf-Attribute:
      case 1: {     // horizontal verschobene Vorzeichen
        int iNoteHead = glb_extraData[i++];
        int alterationShift = (int)(char)glb_extraData[i++];
        break;}
      case 2: {     // Vorzeichen in Klammern
        int iNoteHead = glb_extraData[i++];
        break;}
      // Attribute des gesamten Akkords:
      case 3: break; // Punktierung unterdrckt
      case 4: break; // Balkenuntergliederung
#ifdef _VER6_EXT_
      case 5: {      // Tempowechsel
        assert(nBytes % 4 == 0);
        int nTempi = nBytes / 4;
        union {
          float fTempo;
          BYTE  b[4];
        };
        for (int t=0; t<nTempi; t++) {
          for (int j=0; j<4; j++)
            b[j] = glb_extraData[i++];
          float tempo = fTempo;
        }
        break;}
#endif
      default: // fr weitere zuknftige Erweiterungen
        i += nBytes;
      }
    }
    assert(glb_extraData[i++] == 1);
    assert(i == glb_nExtraBytes);
  }
#endif
}

DEFINITION(Clef) {
  Byte          b;
                                    enum FORM {FORM_G, FORM_C, FORM_F, FORM_PERCUSSION,
                                               FORM_NULL, CLEF_UNCHANGED};
                                    enum LINE {LINE_5, LINE_4, LINE_3, LINE_2, LINE_1};
                                    enum OCT  {OCT_ALTA, OCT_NULL, OCT_BASSA};

                                    FORM form = (FORM) (b & 7);
                                    assert(form <= FORM_NULL);

                                    LINE line = (LINE) ((b >> 3) & 7);
                                    assert(line <= LINE_1);

                                    OCT  oct  = (OCT)  (b >> 6);
                                    assert(oct <= OCT_BASSA);
}

DEFINITION(DrawObjArray) {
  Unsigned      nObjs;              size = nObjs; // Komponente von DrawObjArray
  FOR (nObjs) {
    Byte        type;
    switch (type) {
    case  0: { GroupObj        obj; } break;
    case  1: { TransposableObj obj; } break;
    case  2: { MetafileObj     obj; } break;
    case  3: { SimpleTextObj   obj; } break;
    case  4: { TextObj         obj; } break;
    case  5: { RectEllipseObj  obj; } break;
    case  6: { LineObj         obj; } break;
    case  7: { PolygonObj      obj; } break;
    case  8: { WavyLineObj     obj; } break;
    case  9: { SlurObj         obj; } break;
    case 10: { NotelinesObj    obj; } break;
    case 11: { WedgeObj        obj; } break;
    case 12: { VoltaObj        obj; } break;
    case 13: { BracketObj      obj; } break;
    case 14: { GuitarObj       obj; } break;
    case 15: { TrillObj        obj; } break; default: assert(false);
    }
  }
}

DEFINITION(ExplicitBarline) {
                                    enum {BAR_SINGLE, BAR_DOUBLE, BAR_END,
                                          BAR_REPEND, BAR_REPSTART, BAR_REPENDSTART};
  Byte          b;                  int type = b & 0xF;
                                    assert(type <= BAR_REPENDSTART);
                                    int barMode = b >> 4;
                                    assert(barMode <= 2);
                                    // 0 = auto, 1 = nur Zeilen, 2 = durchgezogen
}

DEFINITION(GuitarObj) {
  BasicDrawObj  base;
  Point         relPos;
  Color         color;
  Word          flags;
  DWord         strings;            // 8 Saiten in 8 Halbbytes
}

DEFINITION(GroupObj) {
  BasicDrawObj  base;
  Point         relPos;
  DrawObjArray  components;
}

DEFINITION(Key) {
  Byte          b;                  int signature = (int)b - 7;
                                    assert(signature >= -7
                                        && signature <= 7); // 7b .. 7#
}

DEFINITION(LineObj) {
  BasicDrawObj  base;
  Point         pt1;
  Point         pt2;
  Color         color;
  Byte          lineWidth;
}

#ifdef TRACE
static int s_fontIndex = 0;
#endif

DEFINITION(LogFont) {               // Fonts, die mehrmals vorkommen,
                                    // werden nur einmal gespeichert!
                                    assert(sizeof(LOGFONT) == 60);
  Unsigned      index;
  if (index == 0) {                 // neuer Font (im Fontcache speichern)
                                    LOGFONT lf;
                                    const int fixedSize = sizeof(lf) - LF_FACESIZE;
    CharArray   fixedInfo(fixedSize);
                                    memcpy(&lf, fixedInfo, fixedSize);
    Color       clr;                color = clr; // Komponente von LogFont
    String      s;                  assert(strlen(s) < LF_FACESIZE);
                                    strcpy(lf.lfFaceName, s);
#ifdef TRACE
    printf("Font #%d = %s\n", ++s_fontIndex, s);
#endif
  }
  else {
    // schon gelesener Font (index im Fontcache)
    // ...
    printf("Font #%d\n", index);
  }
}

DEFINITION(MetafileObj) {
  BasicRectObj  rect;
  Unsigned      size;
  CharArray     enhMetaFileBits(size);
}

DEFINITION(Meter) {
  Byte          numerator;
  Byte          d;                  int log2Denom  = (d & 0x7F) - 1;
                                    bool allaBreve = (d & 0x80) != 0;
}

DEFINITION(Notelines) {
  Byte          b;
  switch (b) {
  case 1: break; // Einlinienzeile
  case 2: break; // Standard (5 Linien)
  default: {                        assert(b == 0);
    CharArray   lines(11);          // max 11 Linien (0=keine, 1=durchgez., 2=gestrichelt)
    break;}
  }
}

DEFINITION(NotelinesObj) {
  BasicDrawObj  base;
  Integer       x0;
  Integer       x1;
  Integer       y;
  Color         color;
  Notelines     noteLines;
}

DEFINITION(Point) {
  Integer       x;                  value.x = x;
  Integer       y;                  value.y = y;
}

DEFINITION(PolygonObj) {
  BasicDrawObj  base;
  Unsigned      nPoints;

  FOR (nPoints)
    Point       p;

  Bool          bFilled;
  Byte          lineWidth;
  Color         clrFill;
  Color         clrLine;
}

DEFINITION(RectEllipseObj) {
  LineObj       diagonal;
  Integer       radius;
  Bool          bFilled;
  Color         clrFill;
}

DEFINITION(Rest) {
  BasicDurationalObj base;
  Byte          b;                  bool bMultiMeasures    = (b & 1) != 0;
                                    // wie in capella 2:
                                    bool bVerticalCentered = (b & 2) != 0;
                                    bool bAddVerticalShift = (b & 4) != 0;
                                    // zustzliche Verschiebung:
                                    assert((b & 0xF8) == 0); // bits 3...7 reserviert
  if (bMultiMeasures)
    Unsigned    fullMeasures;       // >0, Mehrtaktpause (Wert=ganze)
                                    // das Byte c (Notenwert) aus BasicDurationalObj
                                    // hat folgende genderte Bedeutung:
                                    // (c & 0x08): "Kirchenpause"
                                    // (c & 0x03): Zifferndarstellung:
                                    //             0=keine, 1=unten 2=oben
  if (bAddVerticalShift)
    Integer     vertShift;

#ifdef _VER6_EXT_
  // Neue Attribute in capella 6 aus den in DEFINITION(Voice) global abgelegten
  // Zusatzinformationen (glb_nExtraBytes, glb_extraData) holen:
  if (glb_nExtraBytes > 0) {
    assert(glb_nExtraBytes > 1);
    assert(glb_extraData[0] == 1); // Kenung fr capella-4-Format
    int i = 1;
    while (i < glb_nExtraBytes-1) {
      BYTE format = glb_extraData[i++];
      BYTE nBytes = glb_extraData[i++];
      switch (format) {
      case 1: {       // Balkenlage
        assert(nBytes == 2);
        char beamShift = glb_extraData[i++];
        char beamSlope = glb_extraData[i++];
        break;}         
      case 5: {      // Tempowechsel
        assert(nBytes % 4 == 0);
        int nTempi = nBytes / 4;
        union {
          float fTempo;
          BYTE  b[4];
        };
        for (int t=0; t<nTempi; t++) {
          for (int j=0; j<4; j++)
            b[j] = glb_extraData[i++];
          float tempo = fTempo;
        }
        }
        break;
      default: // fr weitere zuknftige Erweiterungen
        i += nBytes;
      }
    }
    assert(i == glb_nExtraBytes-1);
  }
#endif
}

DEFINITION(ScoreLayout) {
  Integer       smallLineDist;
  Integer       normalLineDist;
  Integer       topDist;
  Integer       interDist;
  Byte          txtAlign;           // Stimmenbezeichnungen 0=links, 1=zentriert, 2=rechts
  Byte          adjustVert;         // 0=nein, 1=auer letzte Seite, 3=alle Seiten
  Byte          b;                  bool redundantKeys    = (b & 1) != 0;
                                    bool modernDoubleNote = (b & 2) != 0;
                                    assert((b & 0xFC) == 0); // bits 2...7 reserviert
  Bool          bSystemSeparators;
  Integer       nUnnamed;
  LogFont       namesFont;

  // Musterzeilen
  Unsigned      nStaveLayouts;
  FOR (nStaveLayouts)
    StaveLayout staveLayout;

  // Systemklammern:
  Unsigned      nBrackets;          // Anzahl Klammern
  FOR (nBrackets) {
    Integer     from;
    Integer     to;
    Bool        curly;
  }
}

DEFINITION(SimpleTextObj) {        // capella-2-Textobjekt
  BasicDrawObj  base;
  Point         relPos;
  Byte          align;
  LogFont       font;
  String        text;
}

DEFINITION(SlurObj) {
  BasicDrawObj  base;
  Point         bezierPoint[4];     // Bezier-Punkte
  Color         color;
  Byte          nEnd;
  Byte          nMid;
  Byte          nDotDist;
  Byte          nDotWidth;
}

DEFINITION(Stave) {
  Char          c;                  assert(c == 'V'); // Kennung
  Meter         defaultMeter;
  Byte          iLayout;
  Integer       topDistX;
  Integer       btmDistX;
  Color         color;
  ExtraInfo     extra;

  // Stimmen
  Unsigned      nVoices;
  FOR (nVoices)
    Voice       voice;
}

DEFINITION(StringArray) {
  Unsigned      nStrings;           size = (int)nStrings; // Komponente von StringArray
  FOR (nStrings)
    String      s;
}

DEFINITION(StaveLayout) {
  Byte          barlineMode;
  Notelines     noteLines;
  Bool          bSmall;
  Integer       topDist;
  Integer       btmDist;
  Integer       groupDist;
  Byte          barlineFrom;
  Byte          barlineTo;
  Clef          clef;

  // Schlagzeuginformation
  Byte          b;                  bool bPercussion = (b & 1) != 0;
                                    // Schlagzeugkanal verwenden
                                    bool bSoundMapIn = (b & 2) != 0;
                                    bool bSoundMapOut =(b & 4) != 0;
  if (bSoundMapIn) {                // Umleitungstabelle fr Eingabe vom Keyboard
    Byte        iMin;
    Byte        n;                  assert(n > 0 && iMin + n <= 128);
    CharArray   soundMapIn(n);      // Tabelle fr MIDI-Tne iMin...iMin+n-1
  }
  if (bSoundMapOut) {               // Umleitungstabelle fr das Vorspielen
    Byte        iMin;
    Byte        n;                  assert(n > 0 && iMin + n <= 128);
    CharArray   soundMapOut(n);     // Tabelle fr MIDI-Tne iMin...iMin+n-1
  }

  Integer       sound;
  Integer       volume;
  Integer       transp;

  String        descr;
  String        name;
  String        abbrev;
  String        intermediateName;
  String        intermediateAbbrev;
}

DEFINITION(System) {
  Char          c;                  assert(c == 'S'); // Kennung
  Integer       nAddBarCount;
  Bool          bBarCountReset;
  Char          explLeftIndent;     // < 0 --> Einrckung gem Stimmenbezeichnungen
                                    // >=  --> explizite Einrckung
  Byte          beamMode;
  Unsigned      tempo;
  Color         color;              // fr Systemklammern
    
#ifdef _VER6_EXT_    
  // neu in Version 6.0:
  // Genaueres Systemtempo (float)
  Unsigned      nExtraBytes;
  if (nExtraBytes > 0) {            int nRead = 0;
    while (nRead < nExtraBytes) {
      Byte      type;               nRead++;
      if (type == 1) {              // einziges in Version 6 bekanntes Format:
                                    // Tempo (auch mit Wiederholungen)
        Byte    n;                  nRead++;
                                    assert(n % 4 == 0);
        int nTempi = n / 4;
        union {
          float fTempo;
          BYTE  b[4];
        };
        for (int i=0; i<nTempi; i++) {
          for (int j=0; j<4; j++) {
            Byte part;
            b[j] = part;
          }
          float newTempo = fTempo;
        }
                                    nRead += n;
      }
      else {                        // in Version 4 unbekanntes Format
        Byte n;                     nRead++;
        FOR (n)
          Byte unknown(n);          nRead += n;
      }
      assert(nRead <= nExtraBytes);
    }
    if (nRead < nExtraBytes) {
      int d = nExtraBytes-nRead;
      FOR (d)
        Byte    unknown;
    }
  }
#else
  ExtraInfo     extra;
#endif
    
  Byte          b;                  // bit 0 irrelevant. Wert unbestimmt
                                    bool bJustified    = (b & 2) != 0;
                                    // Randausgleich (Blocksatz)
                                    bool bPageBreak    = (b & 4) != 0;
                                    // nach dem System neue Seite anfangen
                                    int  instrNotation = (b >> 3) & 7;
                                    // 0 = keine Instrumentenbezeichnung
                                    // 1 = abgekrzt, 2 = vollstndig

  Unsigned      nStaves;
  FOR (nStaves)
    Stave       stave;              // Notenzeilen
}

DEFINITION(TextObj) {              // Text im RTF-Format
  BasicRectObj  rect;
  Unsigned      size;
  CharArray     rtfText(size);      // Text ohne abschlieendes '\0'
}

DEFINITION(TransposableObj) {
  BasicDrawObj  base;
  Point         relPos;
  Byte          b;                  assert(b == 12 || b == 21);
  DrawObjArray  variants;           assert(variants.Size() == b);
  Integer       nRefNote;
}

DEFINITION(TrillObj) {
  BasicDrawObj  base;
  Integer       x0;
  Integer       x1;
  Integer       y;
  Color         color;
  Bool          trillSign;
}

DEFINITION(Voice) {
                                    enum {T_REST, T_CHORD, T_CLEF, T_KEY, T_METER,
                                          T_EXPL_BARLINE, T_IMPL_BARLINE, T_PAGE_BKGR};

  Char          c;                  assert(c == 'C'); // Kennung
  Byte          y0Lyrics;
  Byte          dyLyrics;
  LogFont       lyricsFont;
  Byte          stemDir;
  ExtraInfo     extra;

  Unsigned      nNoteObjs;          // Notenobjekte
  FOR (nNoteObjs) {
    Byte        type;
#ifdef _VER_3_
    ExtraInfo   extra;
#else
    Unsigned    nExtra; 
    glb_nExtraBytes = nExtra;
    for (int i=0; i<nExtra; i++) {
      Byte      extra;
      glb_extraData[i] = extra;
    }
#endif
    if (type != T_REST && type != T_CHORD && type != T_PAGE_BKGR)
      Color     color;
    // Die anderen Objekttypen haben eine komprimierte Farbcodierung
    switch (type) {
    case T_REST        : { Rest     rest;      } break;
    case T_CHORD       :
    case T_PAGE_BKGR   : { ChordObj chord;     } break;
    case T_CLEF        : { Clef     clef;      } break;
    case T_KEY         : { Key      key;       } break;
    case T_METER       : { Meter    meter;     } break;
    case T_EXPL_BARLINE: { ExplicitBarline bar;} break;
                                    default: assert(false);
    }
  }
}

DEFINITION(VoltaObj) {
  BasicDrawObj  base;
  Integer       x0;
  Integer       x1;
  Integer       y;
  Color         color;
  Byte          flags;              bool bLeft      = (flags & 1) != 0; // links abgeknickt
                                    bool bRight     = (flags & 2) != 0; // rechts abgeknickt
                                    bool bDotted    = (flags & 4) != 0;
                                    bool allNumbers = (flags & 8) != 0;
  Byte          numbers;            int from = numbers & 0x0F;
                                    int to = (numbers >> 4) & 0x0F;
}

DEFINITION(WavyLineObj) {
  LineObj       line;
  Byte          waveLen;
  Bool          adapt;
}

DEFINITION(WedgeObj) {
  LineObj       line;
  Byte          b;                  int height   = b & 0x7F;
                                    bool decresc = (b & 0x80) != 0;
}
